%  File     : calculator.pl
%  RCS      : $Id: calculator.pl,v 1.1 1998/03/03 01:35:03 pets Exp pets $
%  Author   : Peter Schachte, based on work by Fergus Henderson
%  Origin   : Tue Mar  3 11:08:56 1998
%  Purpose  : Example use of global variables
%  Copyright: (c) 1998 University of Melbourne. \%\
%
% A simpler calculator --- parses and evaluates integer expressions using 
% global variables instead of the more obvious DCG implementation.

:- module(calculator, [
	main/0
   ]).

:- load_files(library(globals), [when(compile_time),if(changed)]).
:- use_module(library(lineio)).
:- use_module(library(ctypes)).

user:runtime_entry(start) :- main.		% define main for Quintus

main ::-
	get_line(Line, Term),
	$line := Line,
	(   is_endfile(Term) ->
		true
	;   expr(Expr),
	    $line = [] ->
		evalexpr(Expr, Num),
		write(Num),
		nl,
		main
	;   write('invalid expression'),
	    nl,
	    main
	).

evalexpr(number(Num), Num).
evalexpr(plus(X,Y),  Z) :- evalexpr(X,A), evalexpr(Y,B), Z is A + B.
evalexpr(minus(X,Y), Z) :- evalexpr(X,A), evalexpr(Y,B), Z is A - B.
evalexpr(times(X,Y), Z) :- evalexpr(X,A), evalexpr(Y,B), Z is A * B.
evalexpr(div(X,Y),   Z) :- evalexpr(X,A), evalexpr(Y,B), Z is A // B.

% Simple recursive-descent parser.

expr(Expr) ::-
	factor(Factor),
	expr_tail(Factor, Expr).


expr_tail(Factor, Expr) ::-
	(   nextchar(C),
	    expr_op(C, Factor, Factor2, Factor3) ->
		factor(Factor2),
		expr_tail(Factor3, Expr)
	;   Expr = Factor
	).

expr_op(0'+, Factor1, Factor2, plus(Factor1, Factor2)).
expr_op(0'-, Factor1, Factor2, minus(Factor1, Factor2)).


factor(Factor) ::-
	term(Term),
	factor_tail(Term, Factor).


factor_tail(Term, Factor) ::-
	(   nextchar(C),
	    factor_op(C, Term, Term2, Term3) ->
		term(Term2),
		factor_tail(Term3, Factor)
	;   Factor = Term
	).

factor_op(0'*, Term1, Term2, times(Term1, Term2)).
factor_op(0'/, Term1, Term2, div(Term1, Term2)).


term(Term) ::-
	(   num(Num) ->
	      Term = number(Num)
	;   nextchar(0'(),
	    expr(Term),
	    nextchar(0'))
	).

num(Num) ::-
	nextchar(Digit),
	is_digit(Digit, 10, Value0),
	num_tail(Value0, Num).

num_tail(Value0, Value) ::-
	(   nextchar(Digit),
	    is_digit(Digit, 10, Value1),
	    Value2 is Value0*10+Value1,
	    num_tail(Value2, Value)
	;   Value = Value0
	).

nextchar(C) ::-
	$line = [C|Tail],		% this could instead be written
	$line := Tail.			% $line = [C | $:=line]
